/* Pipeline is implemented as a wrapper around the more general Pipelinev.  */
/* This is not necessarily optimal.                                         */
void Pipeline (int ns, Imode_t imode[], eSkel_molecule_t * (*stages[])(eSkel_molecule_t *), int col,
               Dmode_t dmode, spread_t spr[], MPI_Datatype ty[], void *in,
               int inlen, int inmul, void *out, int outlen, int *outmul,
               int outbuffsz, MPI_Comm comm)
{
  int *inlens;
  int *outlens;
  int i;

  /* inlens is the array of the length of the input atoms. Pipeline assumes that all these */
  /* length are identical and equal to inlen.                                              */
  inlens = (int *) malloc (sizeof(int)*inmul);
  for (i=0; i<inmul; i++) {
    inlens[i] = inlen;
  }
  
   /* outlens is used to pass some additional parameters to Pipelinev:
     - outlens[0] is the length, in units big enough for ty[ns], of the out buffer
     - outlens[1] is the length, in units big enough for an int, of the outlens buffer
     It is also used later in Pipelinev to put some information, so we need to allocate
     some memory in the case the outlen is non null.
  */
  if (outlen && (outbuffsz/outlen > 1)) {
    outlens = (int *) malloc (sizeof(int)* outbuffsz/outlen);
  }
  else {
    outlens = (int *) malloc (sizeof(int)* 2);
  }
  outlens[0] = outbuffsz;
  outlens[1] = outlen ? outbuffsz/outlen : 0;

  /* Call Pipelinev with the new inlens and outlens */
  Pipelinev (ns, imode, stages, col, dmode, spr, ty, in, inlens, inmul, out, outlens, outmul, comm);
}

