int DEBUGPV=0; // set to 1 for debug information

/* General function for Pipeline */
void Pipelinev (int ns, Imode_t imode[], eSkel_molecule_t * (*stages[])(eSkel_molecule_t *), int col, 
                Dmode_t dmode, spread_t spr[], MPI_Datatype ty[], void *in, int *inlens, int inmul,
                void *out, int *outlens, int *outmul, MPI_Comm comm)
{
  /* outlens[0] is the length, in units big enough for ty[ns], of the out buffer,    */
  /* ie what is the maximum accumulated output length we can handle                  */
  /* outlens[1] is the length, in units big enough for an int, of the outlens buffer */
  /* ie what is the maximum outmult we can handle                                    */

  int i, ROOT;
  int *givelen;
  eSkel_molecule_t *tempitemin;
  eSkel_molecule_t *tempitemout;
  int **len;
  MPI_Group stagerootonlygroup, tempgroup;

  if (DEBUGPV) printf("Entering Pipelinev, rank %d, col %d, Top %d -- ", myrank(), col, Top+1);
  
  Top++;

  // Some of this copying is probably redundant
  FrStack[Top]             = (skel_t *) malloc (sizeof(skel_t));
  FrStack[Top]->skel       = Pskel;
  FrStack[Top]->dmode      = dmode;   
  FrStack[Top]->mycol      = col;   
  FrStack[Top]->mycalltree = NULL;    // for now. Used if this turns out to 
                                      // be the innermost call in a BUF rooted nest (LEAF)
  
  // When entering pipeline for the first time, nbgive=nbtake=0
  FrStack[Top]->nbgive = 0;
  FrStack[Top]->nbtake = 0;

  // If this pipeline is in BUF mode, there's no nesting to get the data
  if (dmode==BUF) {
    if (DEBUGPV) printf("BUF mode\n");
    FrStack[Top]->nestlevel = 1;
  }
  // Otherwise increase the nesting level by 1
  else {
    if (DEBUGPV) printf("STRM mode\n");
    FrStack[Top]->nestlevel = FrStack[Top-1]->nestlevel + 1;
  }

  ROOT = Top - FrStack[Top]->nestlevel + 1;  

  // Copy the parameters into the FrStack
  FrStack[Top]->in        = in;
  FrStack[Top]->inmul     = inmul;
  FrStack[Top]->inlens    = inlens; 
  FrStack[Top]->inty      = ty[0];
  FrStack[Top]->inbufspr  = spr[0];
  FrStack[Top]->out       = out;
  FrStack[Top]->outlenP   = &outlens;
  FrStack[Top]->outmulP   = &outmul;
  FrStack[Top]->outty     = ty[ns];
  FrStack[Top]->outbuffsz = outlens[0];
  FrStack[Top]->outbufspr = spr[ns];

  FrStack[Top]->parentcomm = comm;
  MPI_Comm_size(comm, &FrStack[Top]->parentcommsize);
  MPI_Comm_rank(comm, &FrStack[Top]->parentcommrank);
 
  // Copy informations relative to pipeline
  FrStack[Top]->context.p.ns         = ns;
  FrStack[Top]->context.p.spr        = spr;  
  FrStack[Top]->context.p.ty         = ty;   
  FrStack[Top]->context.p.imode      = imode;   
  FrStack[Top]->context.p.itemsz     = (int *) malloc (sizeof(int)*ns+1);
  for (i=0; i<ns+1; i++)
    MPI_Type_size(ty[i], &FrStack[Top]->context.p.itemsz[i]);

  // Size of the first and last item
  FrStack[Top]->initemsz  = FrStack[Top]->context.p.itemsz[0];
  FrStack[Top]->outitemsz = FrStack[Top]->context.p.itemsz[ns];

  // Creating the new communicators of subgroups relating to "col"
  MPI_Comm_split(comm, col, 0, &FrStack[Top]->mycomm);
  MPI_Comm_rank(FrStack[Top]->mycomm, &FrStack[Top]->myrank);
  MPI_Comm_size(FrStack[Top]->mycomm, &FrStack[Top]->mycommsize);

  // Duplicating the parent communicator for other purpose
  MPI_Comm_dup(FrStack[Top]->parentcomm, &(FrStack[Top]->allcomm1));
  MPI_Comm_dup(FrStack[Top]->parentcomm, &(FrStack[Top]->allcomm2));

  // Need to create a comm just with the stage root
  MPI_Comm_group(FrStack[Top]->mycomm, &tempgroup);
  MPI_Group_incl(tempgroup,1,&zero, &stagerootonlygroup);
  MPI_Comm_create(FrStack[Top]->mycomm, stagerootonlygroup, &(FrStack[Top]->stagerootonlycomm));

  // Specify the ranks in the parent communicator (from 0 to parentcommsize)
  FrStack[Top]->allranks = (int *) malloc (sizeof(int)* FrStack[Top]->parentcommsize);
  for (i=0; i<FrStack[Top]->parentcommsize; i++)
    FrStack[Top]->allranks[i] = i;
  
  // work out who's who, MPI_Allgather store the results in allcols
  FrStack[Top]->context.p.allcols = (int *) malloc (sizeof(int)*FrStack[Top]->parentcommsize);
  MPI_Allgather((void *) &FrStack[Top]->mycol, 1, MPI_INT,
		(void *) FrStack[Top]->context.p.allcols, 1, MPI_INT, comm);


  // Run the pipeline
  
  //-------------------- Implicit interaction mode -----------------------//
  // We have to perform all Give/Take necessary to deal with every input
  //      --> Correspond to the eSkel-1 pipeline1for1
  if (imode[col] == IMPL) {
    if (DEBUGPV) printf("Running pipeline %d in IMPL mode\n", FrStack[0]->myrank);
    
    FrStack[Top]->context.p.shutdown = FALSE;
    tempitemin = (eSkel_molecule_t *) malloc (sizeof (eSkel_molecule_t));
    tempitemout = (eSkel_molecule_t *) malloc (sizeof (eSkel_molecule_t));
    len = (int **) malloc (sizeof (int *));

    // Take data, call the stage function and Give the output data if there is some
    while (tempitemin->data = Take(len))  {
      tempitemin->len = *len;
      tempitemout = stages[col](tempitemin);
      if (tempitemout) Give (tempitemout->data, tempitemout->len);
      else Give (NULL, 0);
    }
    
    // Give the shutdown tag if i'm not the last stage
    if (GetMyGroup() != OutBuffGroup(FrStack[Top]->mycalltree,ROOT)) {
      givelen = (int *) malloc (sizeof(int));
      givelen[0] = SHUTDOWNLEN;
      Give(NULL, givelen);
    }    
  }

  //-------------------- Explicit interaction mode -----------------------//
  // The activities are persistent.
  // We call the stage once and just need to be careful for the shutdown
  else if (imode[col] == EXPL) {
    FrStack[Top]->context.p.shutdown = FALSE;

    if (DEBUGPV) printf("Running pipeline %d in EXPL mode\n", FrStack[0]->myrank);
    stages[col](NULL); 

    // Still Give the shutdown tag if i'm not the last stage
    if  (GetMyGroup() != OutBuffGroup(FrStack[Top]->mycalltree,ROOT)) {
      givelen = (int *) malloc (sizeof(int));
      givelen[0] = SHUTDOWNLEN;
      Give(NULL, givelen); 
    }
  }

  //-------------------- Devolved interaction mode -----------------------//
  // The activity is delegated to a persistently nested skeleton.
  // We just call the stage once, which will start this other skeleton
  else {
    if (DEBUGPV) printf("Running pipeline %d in DEV mode \n", FrStack[0]->myrank);
    stages[col](NULL);
  }

  
  Top--; // and give back the space to be tidy
  return;
}

